/*
 * Copyright (c) 2001,2002 Andrew McGill and Leading Edge Business Solutions
 *
 * redirproxy - redirect (maybe) and run scripts - part of fakepop package.
 *
 * fakepop is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * fakepop is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * fakepop; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: redirproxy.c,v 1.2 2004/02/13 11:30:11 andrewm Exp $ */

#define LINUX_NETFILTER 1
#define BUFSIZE 8192 /* 8kb send/receive buffer */
#ifndef VERSION
	#define VERSION whatever
#endif

#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <syslog.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <getopt.h>
#include <time.h>

/* Transparent proxy */
#if LINUX_NETFILTER
#include <linux/netfilter_ipv4.h>
#endif

extern int errno;		/* just in case */
extern char **environ;		/* for execve */
unsigned long bytessent, bytesrecv;	/* bytes sent and received by client */

/* Log an error message */
void
redir_logerror(char *message)
{
       syslog(LOG_NOTICE, "%s: %s", message, strerror(errno));
}


/* Did we read from client? Then don't try to read again until
 * until we have written it to the server. Did we read from
 * server? Then don't try to read again until we have written to
 * client. */
static void
redir_relay(int clientfd, int serverfd)
{
	unsigned char clientbuffer[BUFSIZE]; /* received from client */
	unsigned char serverbuffer[BUFSIZE]; /* received from server */
	int clientbuflen=0,serverbuflen=0;
	int clientoffset=0,serveroffset=0;
	fd_set rfds,wfds,efds;
	int maxfd;
	int rc;

	maxfd = (clientfd > serverfd) ? clientfd : serverfd;

	while (1) {
		FD_ZERO(&rfds);
		FD_ZERO(&wfds);
		FD_ZERO(&efds);
		if (clientbuflen==0) { FD_SET(clientfd, &rfds); }
		else { FD_SET(serverfd, &wfds); }
		if (serverbuflen==0) { FD_SET(serverfd, &rfds); }
		else { FD_SET(clientfd, &wfds); }
		FD_SET(clientfd, &efds);
		FD_SET(serverfd, &efds);

		switch (select(maxfd + 1, &rfds, &wfds, &efds, NULL)) {
			case -1: redir_logerror("select"); /* some or other error */
			case 0: return;
		}
		if (FD_ISSET(clientfd,&efds)) { /* client error */
			redir_logerror("client error");
			return;
		}
		if (FD_ISSET(serverfd,&efds)) { /* server error */
			redir_logerror("server error");
			return;
		}

		/* Read first */
		if (FD_ISSET(clientfd, &rfds)) {
			clientbuflen = recv(clientfd, clientbuffer, sizeof(clientbuffer), MSG_DONTWAIT);
			clientoffset = 0;
			switch (clientbuflen) {
				case -1: redir_logerror("client recv");
				case 0:  return;
			}
		}
		if (FD_ISSET(serverfd, &rfds)) {
			serverbuflen = recv(serverfd, serverbuffer, sizeof(serverbuffer), MSG_DONTWAIT);
			serveroffset = 0;
			switch (serverbuflen) {
				case -1: redir_logerror("server recv");
				case 0:  return;
			}
		}
		/* Now write */
		if (FD_ISSET(clientfd, &wfds)) {
			rc=send(clientfd, serverbuffer+serveroffset,
				serverbuflen-serveroffset, MSG_DONTWAIT);
			if (rc==-1) {
				redir_logerror("send to client");
				return;
			}
			serveroffset += rc;
			bytesrecv += rc;
			if (serveroffset >= serverbuflen) { serverbuflen =0; }
		}
		if (FD_ISSET(serverfd, &wfds)) {
			rc=send(serverfd, clientbuffer+clientoffset,
				clientbuflen-clientoffset, MSG_DONTWAIT);
			if (rc==-1) {
				redir_logerror("send to server");
				return;
			}
			clientoffset += rc;
			bytessent += rc;
			if (clientoffset >= clientbuflen) { clientbuflen =0; }
		}
	}
}

/* Set an environment variable printf style */
void setenvf(char *variable,char *format, ...)
{
	char buffer[512];
	va_list ap;
	va_start(ap, format);
	if (vsnprintf(buffer, sizeof(buffer), format, ap)==0) {
		setenv(variable,buffer,1);
	}
	va_end(ap);
}

void
redir_logandsetupenv(struct sockaddr_in *client_addr,
	struct sockaddr_in *my_addr,
	struct sockaddr_in *server_addr)
{
	char *client, *server, *my;

	client = strdup(inet_ntoa(client_addr->sin_addr));
	server = strdup(inet_ntoa(server_addr->sin_addr));
	my     = strdup(inet_ntoa(my_addr->sin_addr));
	/* Log the connection and run a script to take note of it */
	syslog(LOG_NOTICE,"Diverting for %s:%d from %s:%d to %s:%d",
		client, (int)ntohs(client_addr->sin_port),
		my,     (int)ntohs(my_addr->sin_port),
		server, (int)ntohs(server_addr->sin_port));
	free(client);
	free(server);
	free(my);

	/* Log the connection and run a script to take note of it */
	setenvf("MY_PORT","%d",     (int)ntohs(my_addr->sin_port));
	setenvf("CLIENT_PORT","%d", (int)ntohs(client_addr->sin_port));
	setenvf("SERVER_PORT","%d", (int)ntohs(server_addr->sin_port));
	setenvf("MY_IP","%s",     inet_ntoa(my_addr->sin_addr));
	setenvf("CLIENT_IP","%s", inet_ntoa(client_addr->sin_addr));
	setenvf("SERVER_IP","%s", inet_ntoa(server_addr->sin_addr));
}

char *cmdbefore, *cmdafter, *cmdunreachable;

int
main(int argc, char *argv[])
{
	size_t sock_sz;
	struct sockaddr_in server_addr, client_addr, my_addr;
	int clientfd, serverfd;
	int true = 1;
	time_t timestart, timeend;
	int verbose = 0;

	/* Command option processing */
	while (1) {
		int c;
		int option_index = 0;
		static struct option long_options[] = {
		    {"before", 1, 0, 'b'},
		    {"after", 1, 0, 'a'},
		    {"unreachable", 1, 0, 'u'},
		    {"verbose", 0, 0, 'v'},
		    {0, 0, 0, 0}
		};

		if ((c = getopt_long (argc, argv, "b:a:u:v",
			long_options, &option_index))==-1) {
			break;
		}
		switch (c) {
		case 'b': /* before - pre script */
			cmdbefore=optarg;
			break;
		case 'a': /* after - post script */
			cmdafter=optarg;
			break;
		case 'u': /* exec command in case of unreachable */
			cmdunreachable=optarg;
			break;
		case 'v': /* show banner */
			verbose=1;
			break;
		default:
			fprintf (stderr, "Unknown argument %c\n", c);
			break;
		}
	}

    	/* now do whatever it is that we are going to do */

	openlog("redirproxy", LOG_PID, LOG_DAEMON);
	clientfd=fileno(stdin);
	if (isatty(clientfd)) {
		fprintf(stdout,"Usage: %s [options]\n"
			"\n"
			"\t-b,--before script     Script to run before connection\n"
			"\t-a,--after script      Script to run after connection\n"
			"\t-u,--unreachable file  File to exec for unreachable connections\n"
			"\n"
			"This application is meant to be called by inetd after redirecting\n"
			"a connection using iptables/ipchains:\n"
			"\tiptables -I PREROUTING -p tcp -p 110 -j DNAT --to 10.0.0.1:20110\n"
			"", argv[0]);
		return 1;
	}
	if ((serverfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		redir_logerror("socket");
		return 1;
	}
	sock_sz = sizeof(server_addr);
	server_addr.sin_family = AF_INET;
	if (getpeername(clientfd, (struct sockaddr *) &client_addr, &sock_sz) != 0) {
		redir_logerror("getpeername cannot get client address");
	}
	if (getsockname(clientfd, (struct sockaddr *) &my_addr, &sock_sz) != 0) {
		redir_logerror("getsockname cannot get intercept address");
	}
	if (getsockopt(clientfd, SOL_IP, SO_ORIGINAL_DST, &server_addr, &sock_sz) != 0) {
		redir_logerror("getsockopt cannot get original destination");
		return 1;
	}
	if ( (memcmp(&server_addr.sin_addr, &my_addr.sin_addr, sizeof(server_addr.sin_addr))==0) &&
		(server_addr.sin_port == my_addr.sin_port) ) {
		fprintf(stderr,"not a redirected connection\n");
		syslog(LOG_NOTICE,"not a redirected connection");
		return 1;
	}

	redir_logandsetupenv(&client_addr, &my_addr, &server_addr);
	if (verbose) {
		fprintf(stderr,"You are being diverted ...\r\n");
		fflush(stderr);
	}

	/* Try to connect, but divert if we fail */
	if (connect(serverfd, (struct sockaddr *) &server_addr, sizeof(server_addr)) < 0) {
		redir_logerror("connect");
		if (cmdunreachable) {
			char *argv[4];
			argv[0] = cmdunreachable;
			argv[1] = NULL;
			syslog(LOG_NOTICE,"Running %s",cmdunreachable);
			execve(cmdunreachable,argv,environ);
			redir_logerror("execve");
			return 1;
		}
		return 1;
	}
	if (cmdbefore) {
		system(cmdbefore);
	}
	setsockopt(clientfd, SOL_TCP, TCP_NODELAY, &true, sizeof(true));
	setsockopt(serverfd, SOL_TCP, TCP_NODELAY, &true, sizeof(true));
	
	time(&timestart);
	redir_relay(clientfd, serverfd);

	/* Connection statistics */
	time(&timeend);
	setenvf("BYTESSENT","%lu", bytessent);
	setenvf("BYTESRECV","%lu", bytesrecv);
	setenvf("BYTES","%lu", bytesrecv+bytessent);
	setenvf("KBYTES","%lu", (bytesrecv+bytessent)/1024);
	setenvf("MBYTES","%lu", (bytesrecv+bytessent)/1024/1024);
	setenvf("SECONDS","%d", (int)(timeend-timestart));

	if (cmdafter) {
		system(cmdafter);
	}

	close(serverfd);
	close(clientfd);
	return 0;
}
